import boto3
from dfmock import DFMock
from moto import mock_s3
import pytest

import s3parq.publish_parq as pub_parq


@pytest.mark.slow
@mock_s3
def test_parquet_sizes():
    """ Running two tests - this needs to work with and without partitions """
    bucket = "testbucket"
    key = "testdataset"
    s3_client = boto3.client('s3')
    s3_client.create_bucket(Bucket=bucket)
    df = DFMock(count=100000)
    df.columns = {"int_col": "int", "str_col": "string",
                  "grouped_col": {"option_count": 2, "option_type": "string"}}
    df.generate_dataframe()
    df.grow_dataframe_to_size(500)
    pub_parq.publish(
        dataframe=df.dataframe, key=key, bucket=bucket, partitions=[])

    for obj in s3_client.list_objects(Bucket=bucket)['Contents']:
        if obj['Key'].endswith(".parquet"):
            assert float(obj['Size']) <= 61 * float(1 << 20)


@pytest.mark.slow
@mock_s3
def test_parquet_sizes_with_partition():
    """ Running two tests - this needs to work with and without partitions """
    bucket = "testbucket"
    key = "testdataset"
    s3_client = boto3.client('s3')
    s3_client.create_bucket(Bucket=bucket)
    df = DFMock(count=100000)
    df.columns = {"int_col": "int", "str_col": "string",
                  "grouped_col": {"option_count": 2, "option_type": "string"}}
    df.generate_dataframe()
    df.grow_dataframe_to_size(500)
    pub_parq.publish(
        dataframe=df.dataframe, key=key, bucket=bucket, partitions=['grouped_col'])

    for obj in s3_client.list_objects(Bucket=bucket)['Contents']:
        if obj['Key'].endswith(".parquet"):
            assert float(obj['Size']) <= 61 * float(1 << 20)



import boto3
from dfmock import DFMock
from mock import patch
from moto import mock_s3
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
import s3fs
import contextlib
import moto

import s3parq.publish_parq as parq
from s3parq.testing_helper import (
    setup_grouped_dataframe,
    setup_random_string,
    sorted_dfs_equal_by_pandas_testing,
    setup_custom_redshift_columns_and_dataframe,
    setup_custom_redshift_columns_and_dataframe_with_null
)


@contextlib.contextmanager
def get_s3_client():
    with moto.mock_s3():
        yield boto3.client('s3')


class Test:
    """ Test class for testing out publish_parq
    Unlike fetch, this functionality is encompassible in one file, with just a
    few sections.
    """

    '''
    Setup helpers
    '''

    def setup_s3(self, s3_client):
        bucket = setup_random_string()
        key = setup_random_string()

        s3_client.create_bucket(Bucket=bucket)

        return bucket, key

    def setup_redshift_params(self):
        redshift_params = {
            'schema_name': 'hamburger_schema',
            'table_name': 'hamburger_table',
            'iam_role': 'hamburger_arn',
            'region': 'us-east-1',
            'cluster_id': 'hamburger_cluster',
            'host': 'hamburger_host',
            'port': '9999',
            'db_name': 'hamburger_db',
            'ec2_user': 'hamburger_aws'
        }

        return redshift_params

    '''
    Internal-facing publish tests
    '''

    def test_reject_protected_name_partitions(self):
        assert parq._check_partition_compatibility("shoe")
        assert parq._check_partition_compatibility("all") is False

    def test_generates_partitions_in_order(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = dataframe.columns[:2]
            with patch('s3parq.publish_parq.pq.write_to_dataset', return_value=None) as mock_method:
                parq._gen_parquet_to_s3(bucket, key, dataframe, partitions)
                arg, kwarg = mock_method.call_args
                assert (kwarg['partition_cols'] == partitions).all()

    def test_df_datatypes(self):
        dataframe = setup_grouped_dataframe()
        assert parq._get_dataframe_datatypes(
            dataframe) == dataframe.dtypes.to_dict()

    def test_partition_datatypes(self):
        dataframe = setup_grouped_dataframe()
        partitions = ["text_col", "int_col", "float_col"]
        dataframe_dtypes = dataframe.dtypes.to_dict()
        part_dtypes = {}
        for part in partitions:
            part_dtypes[part] = dataframe_dtypes.pop(part, None)
        assert parq._get_dataframe_datatypes(
            dataframe, partitions, True) == part_dtypes

    def test_dataframe_sans_partitions(self):
        dataframe = setup_grouped_dataframe()
        partitions = ["text_col", "int_col", "float_col"]
        dataframe_dtypes = dataframe.dtypes.to_dict()
        for part in partitions:
            dataframe_dtypes.pop(part, None)
        assert parq._get_dataframe_datatypes(
            dataframe, partitions) == dataframe_dtypes

    '''
    External publish functionality tests
    '''

    def test_accepts_valid_partitions(self):
        dataframe = setup_grouped_dataframe()
        parq.check_partitions(dataframe.columns, dataframe)

    def test_reject_non_column_partitions(self):
        dataframe = setup_grouped_dataframe()
        with pytest.raises(ValueError):
            parq.check_partitions(('banana',), dataframe)

    def test_reject_empty_dataframe(self):
        with get_s3_client() as s3_client:
            dataframe = pd.DataFrame()
            bucket, key = self.setup_s3(s3_client)
            s3_path = f"s3://{bucket}/{key}"

            with pytest.raises(ValueError):
                parq.publish(bucket=bucket, key=key,
                             dataframe=dataframe, partitions=[])

    def test_reject_timedelta_dataframes(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = ['text_col']
            dataframe['time_col'] = pd.Timedelta('1 days')
            with pytest.raises(NotImplementedError):
                parq.publish(bucket=bucket, key=key,
                             dataframe=dataframe, partitions=partitions)

    def test_works_without_partitions(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions)

    def test_input_equals_output(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            s3_path = f"s3://{bucket}/{key}"
            partitions = [dataframe.columns[0]]
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions)

            from_s3 = pq.ParquetDataset(
                s3_path, filesystem=s3fs.S3FileSystem())
            s3pd = from_s3.read().to_pandas()
            # Switch partition type back -> by default it gets set to a category
            s3pd[partitions[0]] = s3pd[partitions[0]].astype(
                dataframe[partitions[0]].dtype)

            sorted_dfs_equal_by_pandas_testing(dataframe, s3pd)

    def test_set_metadata_correctly(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            s3_client = boto3.client('s3')
            partitions = ['string_col', 'int_col', 'bool_col']
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions)
            for obj in s3_client.list_objects(Bucket=bucket)['Contents']:
                if obj['Key'].endswith(".parquet"):
                    meta = s3_client.get_object(
                        Bucket=bucket, Key=obj['Key'])['Metadata']
                    assert meta['partition_data_types'] == str(
                        {"string_col": "string", "int_col": "integer", "bool_col": "boolean"})

    '''
    External custom publish functionality tests
    '''

    def test_custom_publish_reject_empty_dataframe(self):
        with get_s3_client() as s3_client:
            dataframe = pd.DataFrame()
            custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()[
                1]
            bucket, key = self.setup_s3(s3_client)
            s3_path = f"s3://{bucket}/{key}"

            with pytest.raises(ValueError):
                parq.custom_publish(bucket=bucket, key=key,
                                    dataframe=dataframe, partitions=[],
                                    custom_redshift_columns=custom_redshift_columns)

    def test_custom_publish_reject_timedelta_dataframes(self):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = ['colA']
            dataframe['time_col'] = pd.Timedelta('1 days')
            with pytest.raises(NotImplementedError):
                parq.custom_publish(bucket=bucket, key=key,
                                    dataframe=dataframe, partitions=partitions,
                                    custom_redshift_columns=custom_redshift_columns)

    def test_custom_publish_works_without_partitions(self):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions,
                                custom_redshift_columns=custom_redshift_columns)

    def test_custom_publish_input_equals_output(self):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            s3_path = f"s3://{bucket}/{key}"
            partitions = [dataframe.columns[0]]
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions,
                                custom_redshift_columns=custom_redshift_columns)

            from_s3 = pq.ParquetDataset(
                s3_path, filesystem=s3fs.S3FileSystem())
            s3pd = from_s3.read().to_pandas()
            # Switch partition type back -> by default it gets set to a category
            s3pd[partitions[0]] = s3pd[partitions[0]].astype(
                dataframe[partitions[0]].dtype)

            sorted_dfs_equal_by_pandas_testing(dataframe, s3pd)

    def test_custom_publish_set_metadata_correctly(self):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            s3_client = boto3.client('s3')
            partitions = ['colA', 'colB', 'colC', 'colD', 'colF']
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions,
                                custom_redshift_columns=custom_redshift_columns)
            for obj in s3_client.list_objects(Bucket=bucket)['Contents']:
                if obj['Key'].endswith(".parquet"):
                    meta = s3_client.get_object(
                        Bucket=bucket, Key=obj['Key'])['Metadata']
                    assert meta['partition_data_types'] == str(
                        {"colA": "string", "colB": "integer", "colC": "float", "colD": "decimal", "colF": "boolean"})

    '''
    Tests related to Redshift Spectrum
    '''

    # Test to ensure that exception are raised on missing/wrong params
    def test_validate_redshift_params_wrong_params(self):
        redshift_params = self.setup_redshift_params()
        redshift_params.pop("table_name", None)

        with pytest.raises(ValueError):
            parq.validate_redshift_params(redshift_params)

        redshift_params['tbale_name'] = "some_typoed_mess"
        with pytest.raises(KeyError):
            parq.validate_redshift_params(redshift_params)

    # Test to ensure that redshift parameter validation checks and addresses table name caseing
    def test_validate_redshift_params_table_case(self):
        redshift_params = self.setup_redshift_params()
        redshift_params['table_name'] = 'hAMbURgeR_tAbLE'
        assert parq.validate_redshift_params(redshift_params)[
            'table_name'] == 'hamburger_table'

    # Test to ensure that redshift parameter validation checks and addresses schema name caseing
    def test_validate_redshift_params_schema_case(self):
        redshift_params = self.setup_redshift_params()
        redshift_params['schema_name'] = 'HAmbURger_sCHemA'
        assert parq.validate_redshift_params(redshift_params)[
            'schema_name'] == 'hamburger_schema'

    # Test Schema Creator on Publish
    # TODO : is this necesarry?
    def test_no_redshift_publish(self):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions)

    @patch('s3parq.publish_redshift.create_schema')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_schema_publish(self, mock_session_helper, mock_create_schema):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = [dataframe.columns[0]]
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions, redshift_params=redshift_params)

            mock_create_schema.assert_called_once_with(
                redshift_params['schema_name'], redshift_params['db_name'], redshift_params['iam_role'], msh)

    @patch('s3parq.publish_redshift.create_table')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_table_publish(self, mock_session_helper, mock_create_table):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = ["text_col", "int_col", "float_col"]
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()
            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions, redshift_params=redshift_params)

            df_types = parq._get_dataframe_datatypes(dataframe, partitions)
            partition_types = parq._get_dataframe_datatypes(
                dataframe, partitions, True)

            mock_create_table.assert_called_once_with(
                redshift_params['table_name'], redshift_params['schema_name'], df_types, partition_types, parq.s3_url(bucket, key), msh)

    @patch('s3parq.publish_redshift.create_table')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_table_publish_mixed_type_column(self, mock_session_helper, mock_create_table):
        with get_s3_client() as s3_client:
            dataframe = setup_grouped_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()

            dataframe.iat[5, dataframe.columns.get_loc("text_col")] = 45

            parq.publish(bucket=bucket, key=key,
                         dataframe=dataframe, partitions=partitions, redshift_params=redshift_params)

            df_types = parq._get_dataframe_datatypes(dataframe, partitions)
            partition_types = parq._get_dataframe_datatypes(
                dataframe, partitions, True)

            mock_create_table.assert_called_once_with(
                redshift_params['table_name'], redshift_params['schema_name'], df_types, partition_types, parq.s3_url(bucket, key), msh)

    '''
    Tests related to Redshift Spectrum with custom publish
    '''
    # Test Schema Creator on Publish

    def test_custom_publish_no_redshift_publish(self):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions,
                                custom_redshift_columns=custom_redshift_columns)

    @patch('s3parq.publish_redshift.create_schema')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_custom_publish_schema_publish(self, mock_session_helper, mock_create_schema):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = [dataframe.columns[0]]
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions, redshift_params=redshift_params,
                                custom_redshift_columns=custom_redshift_columns)

            mock_create_schema.assert_called_once_with(
                redshift_params['schema_name'], redshift_params['db_name'], redshift_params['iam_role'], msh)

    @patch('s3parq.publish_redshift.create_custom_table')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_custom_publish_table_publish(self, mock_session_helper, mock_create_custom_table):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = ["colA", "colB", "colC"]
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()
            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions, redshift_params=redshift_params,
                                custom_redshift_columns=custom_redshift_columns)

            mock_create_custom_table.assert_called_once_with(
                redshift_params['table_name'], redshift_params['schema_name'], partitions, parq.s3_url(bucket, key), custom_redshift_columns, msh)

    @patch('s3parq.publish_redshift.create_custom_table')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_custom_table_publish_mixed_type_column(self, mock_session_helper, mock_create_custom_table):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()

            dataframe.iat[1, dataframe.columns.get_loc("colA")] = 45

            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions, redshift_params=redshift_params,
                                custom_redshift_columns=custom_redshift_columns)

            mock_create_custom_table.assert_called_once_with(
                redshift_params['table_name'], redshift_params['schema_name'], partitions, parq.s3_url(bucket, key), custom_redshift_columns, msh)

    @patch('s3parq.publish_redshift.create_custom_table')
    @patch('s3parq.publish_parq.SessionHelper')
    def test_custom_table_publish_null_in_int_column(self, mock_session_helper, mock_create_custom_table):
        with get_s3_client() as s3_client:
            dataframe, custom_redshift_columns = setup_custom_redshift_columns_and_dataframe_with_null()
            bucket, key = self.setup_s3(s3_client)
            partitions = []
            redshift_params = self.setup_redshift_params()
            msh = mock_session_helper(
                region=redshift_params['region'],
                cluster_id=redshift_params['cluster_id'],
                host=redshift_params['host'],
                port=redshift_params['port'],
                db_name=redshift_params['db_name']
            )

            msh.configure_session_helper()

            parq.custom_publish(bucket=bucket, key=key,
                                dataframe=dataframe, partitions=partitions, redshift_params=redshift_params,
                                custom_redshift_columns=custom_redshift_columns)

            mock_create_custom_table.assert_called_once_with(
                redshift_params['table_name'], redshift_params['schema_name'], partitions, parq.s3_url(bucket, key), custom_redshift_columns, msh)



import boto3
import dfmock
import moto
import pandas as pd
from pandas.util.testing import assert_frame_equal
import pytest

from s3parq.fetch_parq import fetch
from s3parq.publish_parq import publish
from s3parq.testing_helper import df_equal_by_set, sorted_dfs_equal_by_pandas_testing


@moto.mock_s3
def test_end_to_end():
    # make a sample DF for all the tests
    df = dfmock.DFMock(count=10000)
    df.columns = {"string_options": {"option_count": 4, "option_type": "string"},
                "int_options": {"option_count": 4, "option_type": "int"},
                "datetime_options": {"option_count": 5, "option_type": "datetime"},
                "float_options": {"option_count": 2, "option_type": "float"},
                "metrics": "integer"
                }
    df.generate_dataframe()

    s3_client = boto3.client('s3')
    bucket_name = 'thistestbucket'
    key = 'thisdataset'
    s3_client.create_bucket(Bucket=bucket_name)

    old_df = pd.DataFrame(df.dataframe)

    # pub it
    publish(
        bucket=bucket_name,
        key=key,
        dataframe=old_df,
        partitions=['string_options', 'datetime_options', 'float_options']
    )

    # go get it
    fetched_df = fetch(
        bucket=bucket_name,
        key=key,
        parallel=False
    )

    assert fetched_df.shape == old_df.shape
    assert df_equal_by_set(fetched_df, old_df, old_df.columns)
    sorted_dfs_equal_by_pandas_testing(fetched_df, old_df)




from mock import patch
import pytest

from s3parq import publish_redshift
from s3parq.testing_helper import setup_custom_redshift_columns_and_dataframe


class MockScopeObj():

    def execute(self, schema_string: str):
        pass


def scope_execute_mock(mock_session_helper):
    pass


class Test():

    # Make sure that naming validator throws expected errors
    def test_naming_validator(self):
        response = publish_redshift._validate_name('my string')
        assert not response[0], 'Allowed name to contain spaces'

        response = publish_redshift._validate_name('my_string')
        assert response[0]

        response = publish_redshift._validate_name('WHERE')
        assert not response[0], 'Allowed name to be a reserved SQL keyword'

        response = publish_redshift._validate_name('@my_string')
        assert not response[0], 'Allowed name to start as not an alphanumeric or an underscore'

        response = publish_redshift._validate_name(
            "asdffdsaasdffdsaasdfasdffdsaasdffdsaasd\
            fasdffdsaasdffdsaasdfasdffdsaasdffdsaasdsd\
            ffdsaasdffdsaasdfasdffdsaasdffdsaasdfasdffdsaasdffdsaasdf"
        )
        assert not response[0], f'Allowed name as too long string'

    # Make sure that the reshift-specific validator throes the right errors
    def test_validator(self):
        schema_name_good = "my_string"
        bad_schema_names = ["my string", "#t3rr1bl13_n4m3", "", "select"]
        database_name_good = 'my_database'
        bad_database_names = ['my database',
                              "#t3rr1bl13_n4m3", "", ".", "select"]
        publish_redshift._redshift_name_validator(
            schema_name_good, database_name_good)
        for bad_schema in bad_schema_names:
            with pytest.raises(ValueError):
                publish_redshift._redshift_name_validator(
                    bad_schema, database_name_good)
        for bad_db in bad_database_names:
            with pytest.raises(ValueError):
                publish_redshift._redshift_name_validator(
                    schema_name_good, bad_db)

    # Test that the function is called with the schema name
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_schema(self, mock_session_helper, mock_execute):

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        schema_name = "my_string"
        db_name = "my_database"
        iam_role = "my_iam_role"
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_schema(
                schema_name, db_name, iam_role, mock_session_helper)
            mock_scope.execute.assert_called_once_with(f"CREATE EXTERNAL SCHEMA IF NOT EXISTS {schema_name} \
                FROM DATA CATALOG \
                database '{db_name}' \
                iam_role '{iam_role}';")

    # Test that the function is called with the table name
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_table(self, mock_session_helper, mock_execute):

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        table_name = "my_string"
        schema_name = "my_schema"
        path = "s3://lol"
        columns = {'grouped_col': 'object', 'text_col': 'object',
                   'int_col': 'int64', 'float_col': 'float64'}
        partitions = {'fish': 'object'}

        expected_sql = f'CREATE EXTERNAL TABLE IF NOT EXISTS {schema_name}.{table_name} {columns} \
            PARTITIONED BY {partitions} STORED AS PARQUET \
            LOCATION "{path}";'
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_table(table_name, schema_name, columns,
                                          partitions, path, mock_session_helper)
            assert mock_scope.execute.called_once_with(expected_sql)

    # Test that the function is called with the table name without partitions
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_table_without_partitions(self, mock_session_helper, mock_execute):

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        table_name = "my_string"
        schema_name = "my_schema"
        path = "s3://lol"
        columns = {'grouped_col': 'object', 'text_col': 'object',
                   'int_col': 'int64', 'float_col': 'float64'}
        partitions = {}

        expected_sql = f'CREATE EXTERNAL TABLE IF NOT EXISTS {schema_name}.{table_name} {columns} \
            STORED AS PARQUET \
            LOCATION "{path}";'
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_table(table_name, schema_name, columns,
                                          partitions, path, mock_session_helper)
            assert mock_scope.execute.called_once_with(expected_sql)

    def test_gets_proper_partitions(self):
        test_str = '/some/path/to/data/banana=33/orange=65/apple=abcd/xyz.parquet'
        final_partitions = publish_redshift._get_partitions_for_spectrum(
            test_str)
        assert final_partitions == ['banana=33', 'orange=65', 'apple=abcd']

    def test_gets_no_partitions(self):
        test_str = '/some/path/to/data/xyz.parquet'
        final_partitions = publish_redshift._get_partitions_for_spectrum(
            test_str)
        assert final_partitions == []

    def test_gets_proper_partitions_multiple_slashes(self):
        test_str = '/some/path/to/data//banana=33/orange=65/apple=abcd/xyz.parquet'
        final_partitions = publish_redshift._get_partitions_for_spectrum(
            test_str)
        assert final_partitions == ['banana=33', 'orange=65', 'apple=abcd']

    def test_format_partition_strings(self):
        test_partitions = ['banana=33', 'orange=65', 'apple=abcd']
        final_partitions = publish_redshift._format_partition_strings_for_sql(
            test_partitions)
        assert final_partitions == [
            "banana='33'", "orange='65'", "apple='abcd'"]

    def test_format_partition_strings_no_partitions(self):
        test_partitions = []
        final_partitions = publish_redshift._format_partition_strings_for_sql(
            test_partitions)
        assert final_partitions == []

    def test_index_containing_substring(self):
        test_list = ['abcd', 'efgh=1234', 'ijkl=5678', 'xyz.parquet']
        index = publish_redshift._last_index_containing_substring(
            test_list, '=')
        assert index == 2

    def test_index_containing_substring_no_match(self):
        test_list = ['abcd', 'efgh=1234', 'ijkl=5678']
        index = publish_redshift._last_index_containing_substring(
            test_list, '&')
        assert index == 4

    def test_get_partition_location(self):
        test_filepath = 'path/to/data/apple=abcd/orange=1234/abcd1234.parquet'
        partition_path = publish_redshift._get_partition_location(
            test_filepath)
        assert partition_path == 'path/to/data/apple=abcd/orange=1234'

    def test_get_partition_location_no_partition(self):
        test_filepath = 'path/to/data/abcd1234.parquet'
        with pytest.raises(ValueError):
            partition_path = publish_redshift._get_partition_location(
                test_filepath)

    # Test that the function is called with the table name without partitions
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_partitions(self, mock_session_helper, mock_execute):

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        table_name = "my_table"
        schema_name = "my_schema"
        bucket = "test"
        partitions = ["version", "time"]
        filepath = "something_overgeneric/dataset/version=v2_final_new/time=01-01-69  23:59:07/keysmash.parquet"
        sql_partitions = "(version='v2_final_new', time='01-01-69  23:59:07')"
        path_for_sql = "'s3://test/something_overgeneric/dataset/version=v2_final_new'"

        expected_sql = f"ALTER TABLE {schema_name}.{table_name} \
#             ADD IF NOT EXISTS PARTITION {sql_partitions} \
#             LOCATION {path_for_sql};"
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_partitions(
                bucket, schema_name, table_name, filepath, mock_session_helper)
            assert mock_scope.execute.called_once_with(expected_sql)

    # Test to check that the passed in datatype maps correctly
    def test_datatype_mapper(self):
        columns = {'grouped_col': 'object', 'text_col': 'object',
                   'int_col': 'int64', 'float_col': 'float64'}
        expected = {'grouped_col': 'VARCHAR', 'text_col': 'VARCHAR',
                    'int_col': 'BIGINT', 'float_col': 'FLOAT'}
        sql = ""
        for key, val in expected.items():
            sql += f'{key} {val}, '
        sql = "(" + sql[:-2] + ")"
        actual = publish_redshift._datatype_mapper(columns)
        assert actual == sql

    # Verify function call for custom create table
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_custom_table(self, mock_session_helper, mock_execute):

        custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()[1]

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        table_name = "my_string"
        schema_name = "my_schema"
        path = "s3://lol"
        columns = {'colA': 'VARCHAR(1000)', 'colB': 'BIGINT',
                   'colC': 'REAL', 'coldD': 'DECIMAL(5,4)',
                   'colE': 'VARCHAR', 'colF': 'BOOLEAN'}
        partitions = {'colA': 'VARCHAR(1000)'}

        expected_sql = f'CREATE EXTERNAL TABLE IF NOT EXISTS {schema_name}.{table_name} {columns} \
            PARTITIONED BY {partitions} STORED AS PARQUET \
            LOCATION "{path}";'
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_custom_table(table_name, schema_name,
                                          partitions, path, custom_redshift_columns, mock_session_helper)
            assert mock_scope.execute.called_once_with(expected_sql)

    # Verify function call for custom create table, no partitions
    @patch('s3parq.publish_redshift.SessionHelper')
    @patch('tests.test_publish_redshift.scope_execute_mock')
    def test_create_custom_table_without_partitions(self, mock_session_helper, mock_execute):

        custom_redshift_columns = setup_custom_redshift_columns_and_dataframe()[1]

        mock_execute.return_value = MockScopeObj()
        mock_session_helper.db_session_scope.return_value.__enter__ = scope_execute_mock

        table_name = "my_string"
        schema_name = "my_schema"
        path = "s3://lol"
        columns = {'colA': 'VARCHAR(1000)', 'colB': 'BIGINT',
                   'colC': 'REAL', 'coldD': 'DECIMAL(5,4)',
                   'colE': 'VARCHAR', 'colF': 'BOOLEAN'}
        partitions = {}

        expected_sql = f'CREATE EXTERNAL TABLE IF NOT EXISTS {schema_name}.{table_name} {columns} \
            STORED AS PARQUET \
            LOCATION "{path}";'
        with mock_session_helper.db_session_scope() as mock_scope:
            publish_redshift.create_custom_table(table_name, schema_name,
                                          partitions, path, custom_redshift_columns, mock_session_helper)
            assert mock_scope.execute.called_once_with(expected_sql)



import boto3
import datetime
import dfmock
from mock import patch
import moto
import pandas as pd
import pytest
import random
import contextlib

import s3parq.fetch_parq as fetch_parq
from s3parq.fetch_parq import MissingS3ParqMetadata
from s3parq.testing_helper import (
    sorted_dfs_equal_by_pandas_testing,
    setup_files_list,
    setup_grouped_dataframe,
    setup_nons3parq_parquet,
    setup_partitioned_parquet,
    setup_random_string
)


@contextlib.contextmanager
def get_s3_client():
    with moto.mock_s3():
        yield boto3.client('s3')


""" This test file is meant for testing the "public" iterface of fetch_parq
For the "private" functionality meant for internal testing, please refer
to test_fetch_parq_internal.py
"""

''' Phase 2 tests:
Refer to functionality of fetch_parq that:
    - Compares partitions and filters
    - Creates path(s) that lead to the appropriate files
    - Gets the list of those files to pull down from
'''

# Test that functionality of max pulls correctly


def test_gets_max():
    with get_s3_client() as s3_client:
        key = "safekeyprefixname/safedatasetname"
        bucket = "safebucketname"
        part_types = {
            "int_col": "int",
            "float_col": "float"
        }

        df = setup_grouped_dataframe(count=10, columns=part_types)
        bucket, parquet_paths = setup_partitioned_parquet(
            dataframe=df,
            bucket=bucket,
            key=key,
            partition_data_types={"int_col": "int"},
            s3_client=s3_client
        )

        fetched_max = fetch_parq.get_max_partition_value(
            bucket=bucket, key=key, partition="int_col")

        # Test max of column is max of the fetched partition
        assert df["int_col"].max() == fetched_max

# Test that the max function denies non-numeric partitions
# TODO: add category when it becomes supported


def test_gets_max_denies_text():
    with get_s3_client() as s3_client:
        key = "safekeyprefixname/safedatasetname"
        bucket = "safebucketname"
        part_types = {
            "string_col": "string",
            "bool_col": "bool"
        }
        col_types = dict(part_types)
        col_types["metrics"] = "int"
        df = setup_grouped_dataframe(count=10, columns=col_types)
        bucket, parquet_paths = setup_partitioned_parquet(
            dataframe=df,
            bucket=bucket,
            key=key,
            partition_data_types=part_types,
            s3_client=s3_client
        )

        with pytest.raises(ValueError):
            fetched_max = fetch_parq.get_max_partition_value(
                bucket=bucket, key=key, partition="string_col")

        with pytest.raises(ValueError):
            fetched_max = fetch_parq.get_max_partition_value(
                bucket=bucket, key=key, partition="bool_col")


''' Phase 3 tests:
Refer to functionality of fetch_parq that:
    - Pulls down the parquet files
    - Transforms the files into dataframes
    - Concatenates the dataframes and returns them
'''

# Test that handles full fetch when no files match


def test_fetch_when_none():
    with get_s3_client() as s3_client:
        input_key = "burger-shipment/buns"
        input_bucket = "loadingdock"
        partitions = ["exp-date"]

        part_types = {
            "count": "int",
            "price": "float",
            "exp-date": "str"
        }

        fetched_dtypes = pd.Series(["int64", "float64", "object"], index=[
            "count", "price", "exp-date"])

        input_df = pd.DataFrame({
            "count": [2, 4, 7, 9],
            "price": [2.43, 1.23, 5.76, 3.28],
            "exp-date": ["x", "z", "a", "zz"]
        })

        setup_partitioned_parquet(
            dataframe=input_df,
            bucket=input_bucket,
            key=input_key,
            partition_data_types={"exp-date": "string"},
            s3_client=s3_client)

        filters = [{
            "partition": "exp-date",
            "comparison": "==",
            "values": ["not-there"]
        }]

        fetched = fetch_parq.fetch(
            bucket=input_bucket,
            key=input_key,
            filters=filters,
            parallel=False
        )

        # Testing that DF is empty and has the expected columns+dtypes
        assert fetched.empty
        assert fetched.dtypes.equals(fetched_dtypes)


def test_get_partition_difference_string():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'hamburger'
    rando_values = [setup_random_string() for x in range(10)]
    s3_paths = [
        f"{key}/{partition}={x}/12345.parquet" for x in rando_values[:-1]]

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        with patch("s3parq.fetch_parq._get_partitions_and_types") as _get_partitions_and_types:
            get_all_files_list.return_value = s3_paths
            _get_partitions_and_types.return_value = {
                "hamburger": "string"}

            # partition values not in list values
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values[:-2])
            assert deltas == [rando_values[-2]]
            # list values not in partition values
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values, True)
            assert deltas == [rando_values[-1]]


def test_get_partition_difference_string_when_none():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'hamburger'
    rando_values = [setup_random_string() for x in range(10)]
    rando_values.sort()
    s3_paths = []

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        get_all_files_list.return_value = []

        # values when theres no bucket data
        deltas = fetch_parq.get_diff_partition_values(
            bucket, key, partition, rando_values[:-2])
        deltas.sort()
        assert deltas == []
        # values when theres no bucket data and reversed
        deltas = fetch_parq.get_diff_partition_values(
            bucket, key, partition, rando_values[:-2], True)
        deltas.sort()
        assert deltas == rando_values[:-2]


def test_get_partition_difference_int_when_none():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'hamburger'
    rando_values = [1, 2, 3, 4, 5]
    rando_values.sort()
    s3_paths = []

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        with patch("s3parq.fetch_parq._get_partitions_and_types") as _get_partitions_and_types:
            get_all_files_list.return_value = []
            _get_partitions_and_types.return_value = 777.09

            # values when there is no bucket data
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values[:-2])
            deltas.sort()
            assert deltas == []
            # values when theres no bucket data and reversed
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values[:-2], True)
            deltas.sort()
            assert deltas == rando_values[:-2]


def test_get_partition_difference_int_comparison_none():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'hamburger'
    rando_values = [1, 2, 3, 4, 5]
    rando_values.sort()
    s3_paths = [
        f"{key}/{partition}={x}/12345.parquet" for x in rando_values[:-1]]

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        with patch("s3parq.fetch_parq._get_partitions_and_types") as _get_partitions_and_types:
            get_all_files_list.return_value = s3_paths
            _get_partitions_and_types.return_value = {
                "hamburger": "integer"}

            # values when there is no sent comparisons
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, [])
            deltas.sort()
            assert deltas == rando_values[:-1]
            # values when there is no sent comparisons and reversed
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, [], True)
            deltas.sort()
            assert deltas == []


def test_get_partition_difference_datetime():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'burgertime'
    rando_values = [(datetime.datetime.now() - datetime.timedelta(
        seconds=random.randrange(100 * 24 * 60 * 60))).replace(microsecond=0) for x in range(5)]
    s3_paths = [
        f"{key}/{partition}={x.strftime('%Y-%m-%d %H:%M:%S')}/12345.parquet" for x in rando_values[:-1]]

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        with patch("s3parq.fetch_parq._get_partitions_and_types") as _get_partitions_and_types:
            get_all_files_list.return_value = s3_paths
            _get_partitions_and_types.return_value = {
                "burgertime": "datetime"}

            # partition values not in list values
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values[:-2])
            assert deltas == [rando_values[-2]]

            # list values not in partition values
            deltas = fetch_parq.get_diff_partition_values(
                bucket, key, partition, rando_values, reverse=True)
            assert deltas == [rando_values[-1]]


def test_get_partition_values():
    bucket = 'safebucket'
    key = 'dataset'
    partition = 'burgertime'
    rando_values = [(datetime.datetime.now() - datetime.timedelta(
        seconds=random.randrange(100 * 24 * 60 * 60))).replace(microsecond=0) for x in range(5)]
    s3_paths = [
        f"{key}/{partition}={x.strftime('%Y-%m-%d %H:%M:%S')}/12345.parquet" for x in rando_values]

    with patch("s3parq.fetch_parq.get_all_files_list") as get_all_files_list:
        with patch("s3parq.fetch_parq._get_partitions_and_types") as _get_partitions_and_types:
            get_all_files_list.return_value = s3_paths
            _get_partitions_and_types.return_value = {
                "burgertime": "datetime"}

            all_values = fetch_parq.get_all_partition_values(
                bucket=bucket, key=key, partition=partition)

            assert set(all_values) == set(rando_values)

# Test that the data in the input but not comparison is fetched


def test_fetches_diff():
    with get_s3_client() as s3_client:
        input_key = "burger-shipment/buns"
        input_bucket = "loadingdock"
        comparison_key = "burger-inventory/buns"
        comparison_bucket = "backroom"
        partitions = ["exp-date"]

        part_types = {
            "count": "int",
            "price": "float",
            "exp-date": "string"
        }

        input_df = pd.DataFrame({
            "count": [2, 4, 7, 9, 9],
            "price": [2.43, 1.23, 5.76, 3.28, 4.44],
            "exp-date": ["x", "z", "a", "zz", "l"]
        })
        comparison_df = pd.DataFrame({
            "count": [2, 3, 4, 9],
            "price": [2.43, 4.35, 1.23, 3.28],
            "exp-date": ["x", "y", "z", "zz"]
        })

        setup_partitioned_parquet(
            dataframe=input_df,
            bucket=input_bucket,
            key=input_key,
            partition_data_types={"exp-date": "string"},
            s3_client=s3_client)

        setup_partitioned_parquet(
            dataframe=comparison_df,
            bucket=comparison_bucket,
            key=comparison_key,
            partition_data_types={"exp-date": "string"},
            s3_client=s3_client)

        test_df = pd.DataFrame({
            "count": [7, 9],
            "price": [5.76, 4.44],
            "exp-date": ["a", "l"]
        })

        fetched_diff = fetch_parq.fetch_diff(
            input_bucket=input_bucket,
            input_key=input_key,
            comparison_bucket=comparison_bucket,
            comparison_key=comparison_key,
            partition=partitions[0],
            parallel=False
        )

        assert fetched_diff.shape == test_df.shape
        sorted_dfs_equal_by_pandas_testing(fetched_diff, test_df)

# Test that the dataframes are fetched properly when one side is empty and the other isnt


def test_fetches_diff_none():
    with get_s3_client() as s3_client:
        input_key = "clay/beads"
        input_bucket = "kiln"
        comparison_key = "new-case"
        comparison_bucket = "storefront"
        partitions = ["price"]

        part_types = {
            "count": "int",
            "price": "float"
        }

        input_df = pd.DataFrame({
            "count": [2, 4, 7, 9],
            "price": [2.43, 1.23, 5.76, 3.28]
        })

        s3_client.create_bucket(Bucket=input_bucket)
        s3_client.create_bucket(Bucket=comparison_bucket)

        setup_partitioned_parquet(
            dataframe=input_df,
            bucket=input_bucket,
            key=input_key,
            partition_data_types={"price": "float"},
            s3_client=s3_client)

        fetched_diff = fetch_parq.fetch_diff(
            input_bucket=input_bucket,
            input_key=input_key,
            comparison_bucket=comparison_bucket,
            comparison_key=comparison_key,
            partition=partitions[0],
            parallel=False
        )

        fetched_diff.sort_values(by=['price'], inplace=True)
        input_df.sort_values(by=['price'], inplace=True)

        sorted_dfs_equal_by_pandas_testing(fetched_diff, input_df)

        fetched_diff_reverse = fetch_parq.fetch_diff(
            input_bucket=input_bucket,
            input_key=input_key,
            comparison_bucket=comparison_bucket,
            comparison_key=comparison_key,
            partition=partitions[0],
            reverse=True,
            parallel=False
        )

        assert fetched_diff_reverse.empty

        fetched_diff_reverse_both = fetch_parq.fetch_diff(
            input_bucket=comparison_bucket,
            input_key=comparison_key,
            comparison_bucket=input_bucket,
            comparison_key=input_key,
            partition=partitions[0],
            reverse=True,
            parallel=False
        )

        sorted_dfs_equal_by_pandas_testing(fetched_diff_reverse_both, input_df)

# Test that all files matching key gets listed out


def test_fetch_files_list():
    with get_s3_client() as s3_client:
        uploaded = setup_files_list(count=100, s3_client=s3_client)

        bucket = uploaded['bucket']
        key = uploaded['key']
        test_files = uploaded['files']

        fetched_files = fetch_parq.get_all_files_list(bucket, key)

        test_files_keyed = list(
            map(lambda x: key + "/" + x + ".parquet", test_files))

        test_files_keyed.sort()
        fetched_files.sort()

        assert (test_files_keyed == fetched_files)

# Test that all files matching key get listed out even with pagination


@pytest.mark.slow
def test_fetch_files_list_more_than_1k():
    with get_s3_client() as s3_client:
        uploaded = setup_files_list(count=1500, s3_client=s3_client)

        bucket = uploaded['bucket']
        key = uploaded['key']
        test_files = uploaded['files']

        fetched_files = fetch_parq.get_all_files_list(bucket, key)

        test_files_keyed = list(
            map(lambda x: key + "/" + x + ".parquet", test_files))

        test_files_keyed.sort()
        fetched_files.sort()

        assert (test_files_keyed == fetched_files)


# Test : fetching without metadata
# Test : fetching when metadata not allowed

def test_fetches_nons3parq():
    with get_s3_client() as s3_client:
        input_key = "burger-shipment/buns"
        input_bucket = "loadingdock"

        input_df = pd.DataFrame({
            "count": [2, 4, 7, 9, 9],
            "price": [2.43, 1.23, 5.76, 3.28, 4.44],
            "exp-date": ["x", "z", "a", "zz", "l"]
        })

        s3_key = "burger-shipment/buns"

        setup_nons3parq_parquet(
            dataframe=input_df,
            bucket=input_bucket,
            key=input_key,
            s3_client=s3_client
        )

        fetched_diff = fetch_parq.fetch(
            bucket=input_bucket,
            key=s3_key,
            parallel=False
        )

        assert fetched_diff.shape == input_df.shape
        sorted_dfs_equal_by_pandas_testing(fetched_diff, input_df)

# test when there's multiple parquet splits


@pytest.mark.slow
def test_fetches_nons3parq_large_parquet():
    input_key = "burger-shipment/buns"
    input_bucket = "loadingdock"

    df = dfmock.DFMock(count=100000)
    df.columns = {"string_options": {"option_count": 4, "option_type": "string"},
                  "int_options": {"option_count": 4, "option_type": "int"},
                  "datetime_options": {"option_count": 5, "option_type": "datetime"},
                  "float_options": {"option_count": 2, "option_type": "float"},
                  "metrics": "integer"
                  }

    df.generate_dataframe()
    # This is unfortunately big, but getting it to force a partition doesn't work otherwise
    df.grow_dataframe_to_size(500)

    input_df = pd.DataFrame(df.dataframe)

    s3_client = boto3.client('s3')

    s3_key = "burger-shipment/buns"

    setup_nons3parq_parquet(
        dataframe=input_df,
        bucket=input_bucket,
        key=input_key,
        s3_client=s3_client
    )

    fetched_diff = fetch_parq.fetch(
        bucket=input_bucket,
        key=s3_key,
        parallel=False
    )

    assert fetched_diff.shape == input_df.shape
    sorted_dfs_equal_by_pandas_testing(fetched_diff, input_df)


def test_not_fetches_nons3parq():
    with get_s3_client() as s3_client:
        input_key = "burger-shipment/buns"
        input_bucket = "loadingdock"

        input_df = pd.DataFrame({
            "count": [2, 4, 7, 9, 9],
            "price": [2.43, 1.23, 5.76, 3.28, 4.44],
            "exp-date": ["x", "z", "a", "zz", "l"]
        })

        s3_key = "burger-shipment/buns"

        setup_nons3parq_parquet(
            dataframe=input_df,
            bucket=input_bucket,
            key=input_key,
            s3_client=s3_client
        )

        with pytest.raises(MissingS3ParqMetadata):
            fetched_diff = fetch_parq.fetch(
                bucket=input_bucket,
                key=s3_key,
                parallel=False,
                accept_not_s3parq=False
            )



import boto3
from collections import OrderedDict
import datetime
from mock import patch
import moto
import pandas as pd
import pytest
import contextlib

import s3parq.fetch_parq as fetch_parq
from s3parq.testing_helper import (
    sorted_dfs_equal_by_pandas_testing,
    setup_grouped_dataframe,
    setup_partitioned_parquet
)


@contextlib.contextmanager
def get_s3_client():
    with moto.mock_s3():
        yield boto3.client('s3')


""" This test file is meant for testing the "private" iterface of fetch_parq
for interal testing. For the "public" functionality testing, please refer
to test_fetch_parq.py
"""

''' Phase 1 tests:
Refer to functionality of fetch_parq that:
    - Takes in parameters and checks for validity
    - Fetches partitions and associated data
'''


def test_filter_validation():
    invalid_filters_list = [
        [{
            "comparison": "==",
            "values": ["fake-value"]
        }],
        [{
            "partition": "fake-part",
            "comparison": "&&",
            "value": "some-string"
        }]
    ]

    for inv_fil in invalid_filters_list:
        with pytest.raises(ValueError):
            fetch_parq._validate_filter_rules(inv_fil)

    valid_filters_list = [
        [{
            "partition": "fake-part",
            "comparison": "==",
            "values": ["some-string"]
        }],
        [{
            "partition": "fake-part",
            "comparison": ">",
            "values": [1]
        }]
    ]

    for val_fil in valid_filters_list:
        fetch_parq._validate_filter_rules(val_fil)


# Test that all valid partitions are correctly parsed


def test_get_partitions():
    parts = [
        "fake-key/fil-1=1/fil-2=2/fil-3=str/rngf1.parquet",
        "fake-key/fil-1=1/fil-2=2/fil-3=rng-str/rngf2.parquet",
        "fake-key/fil-1=1/fil-2=4/fil-3=str_rng/rngf3.parquet",
        "fake-key/fil-1=5/fil-2=2/fil-3=str/rngf4.parquet",
        "fake-key/fil-1=5/fil-2=4/fil-3=99/rngf5.parquet"
    ]
    parsed_parts = OrderedDict({
        "fil-1": {"1", "5"},
        "fil-2": {"2", "4"},
        "fil-3": {"str", "rng-str", "str_rng", "99"}
    })

    key = "fake-key"
    test_parsed_part = fetch_parq._parse_partitions_and_values(parts, key)

    assert parsed_parts == test_parsed_part

# Test that if no partitions are present it'll handle without imploding


def test_get_partitions_none():
    parts = [
        "key/rngf1.parc",
        "key/rngf2.parc",
        "key/rngf3.parc",
        "key/rngf4.parc",
        "key/rngf5.parc"
    ]
    parsed_parts = OrderedDict({})

    key = "key"
    test_parsed_part = fetch_parq._parse_partitions_and_values(parts, key)

    assert parsed_parts == test_parsed_part

# Test it setting the correct partition data types


def test_get_data_types_from_s3():
    with get_s3_client() as s3_client:
        bucket, parquet_paths = setup_partitioned_parquet(s3_client=s3_client)

        s3_client = boto3.client('s3')
        files = s3_client.list_objects_v2(Bucket=bucket)
        first_file_key = files["Contents"][0]["Key"]
        partition_metadata = fetch_parq._get_partitions_and_types(
            first_file_key, bucket)

        assert partition_metadata == {
            "string_col": "string",
            "int_col": "integer",
            "float_col": "float",
            "bool_col": "boolean",
            "datetime_col": "datetime"
        }

    # Test that it errors if data types are mismatched


def test_match_argument_data_types():
    part_types = {"fil-3": "string",
                  "fil-2": "int",
                  "fil-1": "float"
                  }

    parsed_parts = OrderedDict({
        "fil-1": set(["1.23", "5.45"]),
        "fil-2": set(["2", "4"]),
        "fil-3": set(["str", "rng-str", "str_rng", "99"])
    })

    test_typed_parts = OrderedDict({
        "fil-1": set([float(1.23), float(5.45)]),
        "fil-2": set([int(2), int(4)]),
        "fil-3": set(["str", "rng-str", "str_rng", "99"])
    })

    typed_parts = fetch_parq._get_partition_value_data_types(
        parsed_parts=parsed_parts,
        part_types=part_types
    )

    assert typed_parts == test_typed_parts

# Test that it handles correct without false positives


def test_mismatch_argument_data_types():
    part_types = {"fil-2": "int",
                  "fil-1": "int"
                  }

    parsed_parts = OrderedDict({
        "fil-1": set(["1.23", "5.45"]),
        "fil-2": set(["2", "4"])
    })

    with pytest.raises(ValueError):
        fetch_parq._get_partition_value_data_types(
            parsed_parts=parsed_parts,
            part_types=part_types
        )


''' Phase 2 tests:
Refer to functionality of fetch_parq that:
    - Compares partitions and filters
    - Creates path(s) that lead to the appropriate files
    - Gets the list of those files to pull down from
'''

# Test that it errors if filters have no matching partitiion


def test_no_part_for_filter():
    part_types1 = {"fil-2": "int",
                   "fil-1": "int"
                   }

    filters1 = [{
        "partition": "fake-partition",
        "comparison": "==",
        "values": ["fake-value"]
    }]

    with pytest.raises(ValueError):
        fetch_parq._validate_matching_filter_data_type(
            part_types1, filters1)

# Test that it filters partitions fully


def test_filter_all_parts():
    typed_parts = OrderedDict({
        "fil-1": set([float(1.23), float(3.9), float(5.45)]),
        "fil-2": set([int(2), int(4)]),
        "fil-3": set(["str", "rng-str", "str_rng", "99"])
    })

    filters = [{
        "partition": "fil-1",
        "comparison": ">",
        "values": [5]
    }, {
        "partition": "fil-2",
        "comparison": "==",
        "values": [2]
    }, {
        "partition": "fil-3",
        "comparison": "==",
        "values": ["str"]
    }]

    fil_paths = [
        'fake-key/fil-1=5.45/fil-2=2/fil-3=str/'
    ]

    key = "fake-key"
    filters = filters

    filter_paths = fetch_parq._get_filtered_key_list(
        typed_parts=typed_parts, filters=filters, key=key)

    filter_paths.sort()
    fil_paths.sort()

    assert filter_paths == fil_paths

# Test that it filters partitions when only some are filtered


def test_filter_some_parts():
    typed_parts = OrderedDict({
        "fil-1": set([float(1.23), float(3.9), float(5.45)]),
        "fil-2": set([int(2), int(4)]),
        "fil-3": set(["str", "rng-str", "str_rng", "99"])
    })

    filters = [{
        "partition": "fil-1",
        "comparison": ">",
        "values": [5]
    }, {
        "partition": "fil-2",
        "comparison": "==",
        "values": [2]
    }]

    fil_paths = [
        'fake-key/fil-1=5.45/fil-2=2/fil-3=99/',
        'fake-key/fil-1=5.45/fil-2=2/fil-3=str/',
        'fake-key/fil-1=5.45/fil-2=2/fil-3=rng-str/',
        'fake-key/fil-1=5.45/fil-2=2/fil-3=str_rng/'
    ]

    key = "fake-key"
    filters = filters

    filter_paths = fetch_parq._get_filtered_key_list(
        typed_parts=typed_parts, key=key, filters=filters)

    filter_paths.sort()
    fil_paths.sort()

    assert filter_paths == fil_paths

# Test that it handles filters ridding everything


def test_filter_to_none():
    typed_parts = OrderedDict({
        "fil-1": set([float(1.23), float(3.9), float(5.45)]),
        "fil-2": set([int(2), int(4)]),
        "fil-3": set(["str", "rng-str", "str_rng", "99"])
    })

    filters = [{
        "partition": "fil-1",
        "comparison": ">",
        "values": [90]
    }, {
        "partition": "fil-2",
        "comparison": "==",
        "values": [23]
    }]

    fil_paths = []

    key = "fake-key"
    filters = filters

    filter_paths = fetch_parq._get_filtered_key_list(
        typed_parts=typed_parts, key=key, filters=filters)

    filter_paths.sort()
    fil_paths.sort()

    assert filter_paths == fil_paths


''' Phase 3 tests:
Refer to functionality of fetch_parq that:
    - Pulls down the parquet files
    - Transforms the files into dataframes
    - Concatenates the dataframes and returns them
'''

# captures from a single parquet path dataset


def test_s3_parquet_to_dataframe():
    with get_s3_client() as s3_client:
        columns = {"string_col": "string",
                   "int_col": "integer",
                   "float_col": "float",
                   "bool_col": "boolean",
                   "datetime_col": "datetime"}

        bucket = "foobucket"
        key = "fookey"

        df = setup_grouped_dataframe(count=10, columns=columns)
        bucket, parquet_paths = setup_partitioned_parquet(
            dataframe=df,
            bucket=bucket,
            key=key,
            partition_data_types={},
            s3_client=s3_client
        )

        first_published_file = parquet_paths[0]
        response = fetch_parq._s3_parquet_to_dataframe(
            bucket=bucket, key=first_published_file, partition_metadata={})

        assert isinstance(response, pd.DataFrame)
        for col in columns.keys():
            assert (col in response.columns)

        assert response.shape == df.shape
        sorted_dfs_equal_by_pandas_testing(response, df)


def test_s3_partitioned_parquet_to_dataframe():
    with get_s3_client() as s3_client:
        partition_types = {"string_col": "string",
                           "int_col": "integer",
                           "float_col": "float",
                           "bool_col": "boolean",
                           "datetime_col": "datetime"}
        columns = dict(partition_types)
        columns["metrics"] = "int"

        bucket = "foobucket"
        key = "fookey"

        df = setup_grouped_dataframe(count=10, columns=columns)
        bucket, parquet_paths = setup_partitioned_parquet(
            dataframe=df,
            bucket=bucket,
            key=key,
            partition_data_types=partition_types,
            s3_client=s3_client
        )

        first_published_file = parquet_paths[0]
        response = fetch_parq._s3_parquet_to_dataframe(
            bucket=bucket, key=first_published_file, partition_metadata=partition_types)

        assert isinstance(response, pd.DataFrame)
        for col in columns.keys():
            assert (col in response.columns)

        full_response = pd.DataFrame()
        for path in parquet_paths:
            full_response = full_response.append(fetch_parq._s3_parquet_to_dataframe(
                bucket=bucket, key=path, partition_metadata=partition_types))

        assert full_response.shape == df.shape
        sorted_dfs_equal_by_pandas_testing(full_response, df)



import pytest

from s3parq.session_helper import SessionHelper


class Test:

    def test_no_ec2(self):
        '''Verify that False is returned if not run on an ec2 server'''
        is_ec2 = SessionHelper._is_ec2(self)
        assert is_ec2 is False

        '''Verify that ec2 flag evaluates to false if not run on an ec2 server'''
        sh = SessionHelper(
            region='cheeseburger',
            cluster_id='cheeseburger_id',
            host='cheeseburger_host',
            port='cheeseburger_port',
            db_name='cheeseburger_params',
            ec2_user='cheeseburger_params')
        assert sh.is_ec2_flag is False

    def test_iam_user(self):
        '''Verify that iam_user will assume the value of ec2_user when run on an ec2 server'''
        sh = SessionHelper(
            region='cheeseburger',
            cluster_id='cheeseburger_id',
            host='cheeseburger_host',
            port='cheeseburger_port',
            db_name='cheeseburger_params',
            ec2_user='cheeseburger_user')
        # Fake is_ec2_flag to be True
        sh.is_ec2_flag = True
        sh.set_iam_user()
        assert sh.iam_user is 'cheeseburger_user'

    def test_no_ec2_user(self):
        '''Verify that SessionHelper can accept None as an argument for ec2 user'''
        sh = SessionHelper(
            region='cheeseburger',
            cluster_id='cheeseburger_id',
            host='cheeseburger_host',
            port='cheeseburger_port',
            db_name='cheeseburger_params',
            ec2_user=None)
        assert(type(sh.ec2_user)) is type(None)

    def test_ec2_user(self):
        '''Verify that SessionHelper can accept a string as an argument for ec2 user'''
        sh = SessionHelper(
            region='cheeseburger',
            cluster_id='cheeseburger_id',
            host='cheeseburger_host',
            port='cheeseburger_port',
            db_name='cheeseburger_params',
            ec2_user='cheeseburger_user')
        assert(isinstance(sh.ec2_user, str)) is True



import pytest
from s3parq.s3_naming_helper import S3NamingHelper
from typing import NamedTuple


class TestCase(NamedTuple):
    name: str
    test_name: str


def test_validate_bucket_name():
    helper = S3NamingHelper()

    invalid_cases = [
        # must be between 3-63 chars
        TestCase(name="ab", test_name="allowed bucket name that was too short"),
        TestCase(name=''.join([str(x) for x in range(0, 65)]),
                 test_name='allowed bucket name that was too long'),

        # lower case chars, numbers, periods, dashes
        TestCase(name='_$Bucket',
                 test_name='allowed bucket name with invalid chars'),

        # cannot end with dash
        TestCase(name="bucket-", test_name='allowed bucket name with dash ending'),

        # cannot end with consecutive periods
        TestCase(name="bucket..",
                 test_name="allowed bucket name with double periods"),

        # dashes next to periods
        TestCase(name="bucket-.",
                 test_name="allowed bucket name with dash next to period"),

        # char or number after period
        TestCase(name="bucket.",
                 test_name="allowed bucket name without a letter or number after period)"),

        TestCase(name="_bucket",
                 test_name="allowed bucket name without a letter or number to start")
    ]

    valid_names = ["bucket"]
    for invalid_case in invalid_cases:
        with pytest.raises(ValueError) as excinfo:
            helper.validate_bucket_name(invalid_case.name)

    for valid_name in valid_names:
        helper.validate_bucket_name(valid_name)


def test_validate_part():
    helper = S3NamingHelper()
    response = helper.validate_part('this/is/invalid', allow_prefix=False)

    assert not response[0], f'allowed prefix when prefix was disallowed'

    response = helper.validate_part('')
    assert not response[0], f'allowed blank part'

    response = helper.validate_part('/abc/$$!badval/def')
    assert not response[0], f'allowed bad compound part'


def test_validate_s3_path():
    helper = S3NamingHelper()

    response = helper.validate_s3_path('abc/not/valid')
    assert not response[0], f'allowed s3 path without arn prefix'

    response = helper.validate_s3_path('s3://%%$Bucket_name/is/bad')
    assert not response[0], f'allowed bad bucket name'

    response = helper.validate_s3_path('s3://bucket/path/B#)$_ad/dataset')
    assert not response[0], f'allowed bad bucket prefix'

    response = helper.validate_s3_path('s3://bucket/path/all/good')
    assert response[0], f'disallowed good s3 path'
